import 'dart:async';

import 'package:example/advanced_tab_example.dart';
import 'package:flutter/material.dart';
import 'package:flutter/services.dart';
import 'package:path_stack/path_stack.dart';

class FullApiExample extends StatefulWidget {
  @override
  _FullApiExampleState createState() => _FullApiExampleState();
}

class _FullApiExampleState extends State<FullApiExample> {
  String currentPath = "/";

  @override
  void initState() {
    super.initState();
    RawKeyboard.instance.addListener(_handleKeyDown);
  }

  void _handleKeyDown(RawKeyEvent value) {
    if (value is RawKeyDownEvent) {
      final k = value.logicalKey;
      if (k == LogicalKeyboardKey.digit1) setState(() => currentPath = "home");
      if (k == LogicalKeyboardKey.digit2) setState(() => currentPath = "settings/"); // Should fallback to /alerts
      if (k == LogicalKeyboardKey.digit3) setState(() => currentPath = "settings/alerts");
      if (k == LogicalKeyboardKey.digit4) setState(() => currentPath = "settings/profile");
      if (k == LogicalKeyboardKey.digit5) setState(() => currentPath = "settings/billing");
      if (k == LogicalKeyboardKey.digit6) setState(() => currentPath = "settings/billing/");
      if (k == LogicalKeyboardKey.digit7) setState(() => currentPath = "settings/billing/?foo=99&bar=88");
      if (k == LogicalKeyboardKey.digit8) setState(() => currentPath = "settings/billing_alt/9/8");
      // This should produce 404 if stack is set to caseSensitive
      if (k == LogicalKeyboardKey.digit9) setState(() => currentPath = "HOME");
      if (k == LogicalKeyboardKey.digit0) setState(() => currentPath = "admin");
    }
  }

  @override
  void dispose() {
    RawKeyboard.instance.removeListener(_handleKeyDown);
    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    bool isLoggedIn = false;
    // Outer stack
    return Scaffold(
      body: Stack(
        children: [
          Text(currentPath),
          PathStack(
            // Path is the source of truth for each stack, usually this is shared by all child stacks
            path: currentPath,
            // Optional: Provide custom widget for unknown paths
            unknownPathBuilder: (_) => Center(child: Text("Custom 404 Page")),
            // Optional: Provide custom animationIn (default is no animation)
            transitionBuilder: (_, stack, anim1) => FadeTransition(opacity: anim1, child: stack),
            // Optional: Provide duration for animationIn
            transitionDuration: Duration(milliseconds: 200),
            // By default pages are case insensitive, like a web server. But you can turn this off
            caseSensitive: true,
            // Define all matching routes for this stack
            routes: {
              ["home", "/"]: HomePage().buildStackRoute(),
              // Example of a route guard, you can return any path you want to intercept the route change
              ["admin"]: AdminPage().buildStackRoute(onBeforeChange: (path) {
                if (isLoggedIn) return true;
                scheduleMicrotask(() => setState(() => currentPath = "home"));
                return false;
              }),
              // Adding a "/" at the end of any path indicates it can match as a prefix
              ["settings/"]: PathStack(
                path: currentPath,
                // Set a basePath that will be combined with all routes in this child stack
                basePath: "settings/",
                // Use scaffoldBuilder to add a shared scaffold to all of this stacks child routes
                scaffoldBuilder: (_, stack) => Column(children: [
                  Padding(padding: EdgeInsets.all(20), child: Text("Shared Settings App Bar")), // Settings Header
                  Expanded(child: stack),
                ]),
                // Define child routes for settings section
                routes: {
                  // By adding a "" alias, the first route will match both `/settings/alerts` and `/settings/`
                  ["alerts", ""]: AlertsPage().buildStackRoute(),
                  ["profile"]: ProfilePage().buildStackRoute(),
                  // When you need the args use the full builder. You can use path format, or query string format to pass args.
                  // Path Format needs to define how many arguments it wants, and their names.
                  ["billing_alt/:foo/:bar"]:
                      StackRouteBuilder(builder: (BuildContext context, Map<String, String> args) {
                    return BillingPage(id: "${args["foo"]}_${args["bar"]}");
                  }),
                  // Query string format does not need to define arg names up front.
                  // This will catch billing/ + queryParams, and by adding a "billing" alias, we'll also catch links without the trailing slash
                  ["billing/", "billing"]: StackRouteBuilder(builder: (BuildContext context, Map<String, String> args) {
                    return BillingPage(id: "${args["foo"]}_${args["bar"]}");
                  }),
                },
              ).buildStackRoute(),
            },
          ),
        ],
      ),
    );
  }
}

String instructions = "Use the num keys to switch views.";

class HomePage extends StatelessWidget {
  @override
  Widget build(BuildContext context) =>
      Container(color: Colors.red, child: Text("HOME, $instructions"), width: double.infinity, height: double.infinity);
}

class AlertsPage extends StatelessWidget {
  @override
  Widget build(BuildContext context) => Container(
      color: Colors.green, child: Text("ALERTS, $instructions"), width: double.infinity, height: double.infinity);
}

class ProfilePage extends StatelessWidget {
  @override
  Widget build(BuildContext context) => Container(
      color: Colors.orange, child: Text("PROFILE, $instructions"), width: double.infinity, height: double.infinity);
}

class AdminPage extends StatelessWidget {
  @override
  Widget build(BuildContext context) => Container(
      color: Colors.orange, child: Text("Protected, $instructions"), width: double.infinity, height: double.infinity);
}

class BillingPage extends StatelessWidget {
  const BillingPage({Key? key, required this.id}) : super(key: key);
  final String id;

  @override
  Widget build(BuildContext context) =>
      Container(color: Colors.purple, child: Text("BILLING, id=$id"), width: double.infinity, height: double.infinity);
}
