import 'package:flutter/material.dart';
import 'package:path_to_regexp/path_to_regexp.dart';

class StackRouteBuilder {
  StackRouteBuilder({required this.builder, this.maintainState = true, this.onBeforeChange});

  /// A builder that can inject args into your views before they are built
  final Widget Function(BuildContext context, Map<String, String> args) builder;

  /// Will be called before this route is shown, you can return any substitute route to perform a redirection
  final bool Function(String newRoute)? onBeforeChange;

  /// Indicates that this route should be cached in memory when it is not on-screen.
  /// This maintains all stateful widgets, textControllers, scrollers etc but consumes more RAM.
  final bool maintainState;
}

extension PathStackRouteExtensions on Widget {
  StackRouteBuilder buildStackRoute({bool maintainState = true, bool Function(String newRoute)? onBeforeChange}) =>
      StackRouteBuilder(builder: (_, __) => this, onBeforeChange: onBeforeChange, maintainState: maintainState);
}

class PathStack extends StatefulWidget {
  static const kDefaultDuration = Duration(milliseconds: 250);

  final String path;
  final String basePath;
  final Map<List<String>, StackRouteBuilder> routes;
  final bool caseSensitive;
  final Widget Function(BuildContext context, Widget child)? scaffoldBuilder;
  final Widget Function(BuildContext context, Widget child, AnimationController animation)? transitionBuilder;
  final Widget Function(BuildContext context)? unknownPathBuilder;
  final Duration transitionDuration;

  PathStack(
      {Key? key,
      required this.path,
      required this.routes,
      // TODO: Would be nice to figure out a way to inherit path automatically, but not sure how that would work.
      this.basePath = "",
      this.caseSensitive = false,
      this.scaffoldBuilder,
      this.transitionBuilder,
      this.transitionDuration = kDefaultDuration,
      this.unknownPathBuilder})
      : super(key: key);

  static PathStackState of(BuildContext context) =>
      (context.dependOnInheritedWidgetOfExactType<_InheritedPathStack>() as _InheritedPathStack).state;

  @override
  PathStackState createState() => PathStackState();
}

class _RouteWithBuilder {
  _RouteWithBuilder(this.path, this.content);
  final String path;
  final StackRouteBuilder content;
}

class PathStackState extends State<PathStack> with SingleTickerProviderStateMixin {
  Map<String, Widget> _knownRoutes = {};
  String? _currentRoute;
  late AnimationController transitionInAnim = AnimationController(vsync: this, duration: widget.transitionDuration)
    ..forward();

  @override
  Widget build(BuildContext context) {
    // Try and find a known route for the current path
    _RouteWithBuilder routeToShow = getRouteForPath(widget.path);
    // Check onBeforeChange,
    bool allowChange = routeToShow.content.onBeforeChange?.call(routeToShow.path) ?? true;

    if (allowChange) {
      // Extract queryString and path args
      Map<String, String> args = _PathUtils.argsFromPath(route: widget.basePath + routeToShow.path, path: widget.path);

      // Cache the latest version of this route (if maintainState = false, we'll remove it later)
      _knownRoutes[routeToShow.path] = _KeyedWidget(
        key: ValueKey(routeToShow.path),
        child: routeToShow.content.builder.call(context, args),
      );

      // Trigger transitionIn anim
      if (_currentRoute != routeToShow.path) {
        transitionInAnim.forward(from: 0);
      }
      _currentRoute = routeToShow.path;
    }
    // Get a list of the children we want to show this frame, and figure out the index of our current route
    List<Widget> children = _knownRoutes.values.toList(); // Make a copy of the current knownChildren
    int index = _knownRoutes.keys.toList().indexWhere((r) => r == _currentRoute);

    if (allowChange) {
      // Remove this from the cache if maintainState = false
      if (routeToShow.content.maintainState == false) {
        _knownRoutes.remove(routeToShow.path);
      }
    }

    // Wrap builders
    Widget content = IndexedStack(index: index, children: children);
    content = widget.transitionBuilder?.call(context, content, transitionInAnim) ?? content;
    content = widget.scaffoldBuilder?.call(context, content) ?? content;
    // Wrap content with inherited widget so we can look it up
    return _InheritedPathStack(child: content, state: this);
  }

  void clearState() => _knownRoutes.clear();

  @override
  void dispose() {
    clearState();
    super.dispose();
  }

  bool checkPathMatchesRoute(String path, List<String> routes) {
    //bool allowSuffix(String value) => value.endsWith("/");
    // The first entry in the list of routes will determine whether it allowsSuffix or not... TODO: revisit this decision after more use-case testing
    bool allowSuffix = routes[0].endsWith("/");
    for (var i = 0; i < routes.length; i++) {
      final regExp =
          pathToRegExp('${widget.basePath}${routes[i]}', prefix: allowSuffix, caseSensitive: widget.caseSensitive);
      if (regExp.hasMatch("$path")) return true;
    }
    return false;
  }

  _RouteWithBuilder getRouteForPath(String path) {
    _RouteWithBuilder? matchingRoute;
    // For each route, check all paths it has defined and see if we have a match
    widget.routes.forEach((key, value) {
      if (checkPathMatchesRoute(path, key) && matchingRoute == null) {
        matchingRoute = _RouteWithBuilder(key.join(","), value);
      }
    });
    // If unable to find route, fallback to unknownPathBuilder or a default 404 page.
    if (matchingRoute == null) {
      print("WARNING: Unable to find a route for $path");
      matchingRoute = _RouteWithBuilder("404", StackRouteBuilder(builder: (c, _) {
        return widget.unknownPathBuilder?.call(c) ??
            Scaffold(
              body: Center(
                child: Column(
                  mainAxisSize: MainAxisSize.min,
                  children: [
                    Text("Page Not Found :(", style: TextStyle(fontSize: 26)),
                    Text("Unable to match a route to: '$path'"),
                  ],
                ),
              ),
            );
      }));
    } else {
      print("RouteMatched, ${matchingRoute?.path}");
    }
    return matchingRoute!;
  }
}

class _PathUtils {
  static Map<String, String> argsFromPath({required String route, required String path}) {
    // Get path params
    List<String> paramList = [];
    final regExp = pathToRegExp(route, parameters: paramList);
    Map<String, String> pathMap = {};
    if (regExp.hasMatch(path)) {
      final match = regExp.matchAsPrefix(path)!;
      pathMap = extract(paramList, match);
    }
    // Get queryString params
    Map<String, String> queryMap = Uri.parse(path).queryParameters;
    // Combine path and query args into one, the query args will override the path items with the same id
    final o = {...pathMap, ...queryMap};
    return o;
  }
}

class _KeyedWidget extends StatelessWidget {
  const _KeyedWidget({Key? key, required this.child}) : super(key: key);
  final Widget child;
  @override
  Widget build(BuildContext context) => child;
}

class _InheritedPathStack extends InheritedWidget {
  _InheritedPathStack({Key? key, required Widget child, required this.state}) : super(key: key, child: child);
  final PathStackState state;
  @override
  bool updateShouldNotify(covariant InheritedWidget oldWidget) => true;
}
